from ipaddress import IPv4Address, IPv4Network


class EnsureVPNException(Exception):
    pass


class VPNNotConnectedException(EnsureVPNException):
    def __init__(self, message: str, actual_ip: IPv4Address = None):
        super().__init__(message)
        self.actual_ip = actual_ip



import os

ENSURE_VPN_VERSION = "0.4.0"
USER_AGENT = f"ensure_vpn-v{ENSURE_VPN_VERSION} github.com/ftruzzi/ensure_vpn/"

HIDEMYASS_CHECKER_URL = "https://my.hidemyass.com/vpnbackend/isInVpnTunnel"
MULLVAD_CHECKER_URL = "https://ipv4.am.i.mullvad.net/json"
NORDVPN_CHECKER_URL = "https://nordvpn.com/wp-admin/admin-ajax.php"
PROTONVPN_SERVER_URL = "https://api.protonmail.ch/vpn/logicals"
PROTONVPN_SERVER_FILE_PATH = os.path.join(
    os.path.dirname(__file__), "proton_servers.json"
)

IP_CHECKERS = [
    "ifconfig.me",
    "icanhazip.com",
    "ipinfo.io/ip",
    "api.ipify.org",
    "ident.me",
]



import abc
import os
import json
import time

from ipaddress import IPv4Address, IPv4Network
from os.path import isfile
from typing import List, Tuple

import requests

from .checkers import APIChecker, EnsureVPNResult, IPChecker
from .constants import (
    HIDEMYASS_CHECKER_URL,
    MULLVAD_CHECKER_URL,
    NORDVPN_CHECKER_URL,
    PROTONVPN_SERVER_URL,
    PROTONVPN_SERVER_FILE_PATH,
    USER_AGENT,
)
from .helpers import get_dict_values, is_today


class VPNProvider(abc.ABC):
    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        Name of the VPN provider
        """
        raise NotImplementedError("Every VPN provider must have a name.")

    @abc.abstractmethod
    def validate(self) -> EnsureVPNResult:
        """
        IP/connection validation function.
        """
        raise NotImplementedError("Every VPN provider must have a validation function.")


class CustomVPN(VPNProvider):
    name = ""
    negated = False

    def __init__(self, wanted_ip: str) -> None:
        self.name = wanted_ip
        self.wanted_ip = IPv4Network(wanted_ip)

    def validate(self) -> EnsureVPNResult:
        checker = IPChecker(
            validation_func=lambda ip: self.wanted_ip.overlaps(IPv4Network(ip))
        )
        return checker.run()


class HideMyAssVPN(VPNProvider):
    name = "HideMyAss"

    @staticmethod
    def validate() -> EnsureVPNResult:
        checker = APIChecker(
            url=HIDEMYASS_CHECKER_URL,
            validation_func=lambda json: json["isInVpnTunnel"] is True,
            ip_func=lambda _: IPChecker._get_current_ip(),
        )
        return checker.run()


class MullvadVPN(VPNProvider):
    name = "Mullvad"

    @staticmethod
    def validate() -> EnsureVPNResult:
        checker = APIChecker(
            url=MULLVAD_CHECKER_URL,
            validation_func=lambda json: json["mullvad_exit_ip"] is True,
            ip_func=lambda json: IPv4Address(json["ip"]),
        )
        return checker.run()


class NordVPN(VPNProvider):
    name = "NordVPN"

    @staticmethod
    def validate() -> EnsureVPNResult:
        checker = APIChecker(
            url=NORDVPN_CHECKER_URL,
            params={"action": "get_user_info_data"},
            validation_func=lambda json: json["status"] is True,
            ip_func=lambda json: IPv4Address(json["ip"]),
        )
        return checker.run()


class ProtonVPN(VPNProvider):
    name = "ProtonVPN"

    @staticmethod
    def _fetch_servers() -> List[str]:
        json = requests.get(
            PROTONVPN_SERVER_URL, headers={"User-Agent": USER_AGENT}
        ).json()
        return list(get_dict_values("ExitIP", json))

    @staticmethod
    def _get_servers(fetch_servers: bool = False) -> Tuple[List[str], bool]:
        """
        Returns ProtonVPN server list from file or network
        """
        if (
            fetch_servers is True
            or not isfile(PROTONVPN_SERVER_FILE_PATH)
            or not is_today(time.ctime(os.path.getmtime(PROTONVPN_SERVER_FILE_PATH)))
        ):
            servers = ProtonVPN._fetch_servers()
            with open(PROTONVPN_SERVER_FILE_PATH, "w") as f:
                json.dump(servers, f)
            return servers, True

        return json.load(open(PROTONVPN_SERVER_FILE_PATH, "r")), False

    @staticmethod
    def validate() -> EnsureVPNResult:
        servers, were_fetched = ProtonVPN._get_servers()
        checker = IPChecker(
            validation_func=lambda ip: str(ip) in servers
        )

        result = checker.run()
        if result.is_connected or were_fetched:
            return result

        # if the check failed with cached server file, retry with new server list
        servers, _ = ProtonVPN._get_servers(fetch_servers=True)
        return checker.run()



import abc

from ipaddress import IPv4Address, IPv4Network
from json import JSONDecodeError
from typing import Any, Callable, Union

import requests
from requests.exceptions import RequestException

from .constants import IP_CHECKERS, USER_AGENT


class EnsureVPNResult:
    def __init__(self, is_connected: bool, actual_ip: Union[str, IPv4Address]):
        self.is_connected = is_connected
        if isinstance(actual_ip, str):
            actual_ip = IPv4Address(actual_ip)
        self.actual_ip = actual_ip


class VPNChecker(abc.ABC):
    @abc.abstractmethod
    def run(self) -> EnsureVPNResult:
        raise NotImplementedError("Every checker must have a `run` method.")


class APIChecker(VPNChecker):
    """
    requests-based checker that performs HTTP requests.
    """

    def __init__(
        self,
        *,
        url: str,
        validation_func: Callable[[Any], bool],
        ip_func: Callable[[Any], IPv4Address],
        **request_args,
    ):
        """requests-based checker that performs HTTP requests.

        Args:
            url (str): URL to perform a GET request to
            request_args (dict): args passed to requests.request call
            validation_func (Callable[[Any], bool]): checks response to validate connection
            ip_func (Callable[[Any], str]): retrieves actual IP from response
        """
        self.url = url
        self.request_args = request_args
        self.validation_func = validation_func
        self.ip_func = ip_func

        self.session = requests.Session()
        self.session.headers.update({"User-Agent": USER_AGENT})

    @staticmethod
    def _get_response(response: requests.Response) -> Union[dict, str]:
        response.raise_for_status()
        try:
            return response.json()
        except JSONDecodeError:
            return response.text

    def run(self) -> EnsureVPNResult:
        """Runs checker, fetching and validating data

        Returns:
            EnsureVPNResult: contains check result and actual IP
        """
        response = APIChecker._get_response(
            self.session.request(method="GET", url=self.url, **self.request_args)
        )

        return EnsureVPNResult(
            is_connected=self.validation_func(response),
            actual_ip=self.ip_func(response),
        )


class IPChecker(VPNChecker):
    def __init__(
        self,
        *,
        validation_func: Callable[[IPv4Address], bool],
    ):
        self.validation_func = validation_func

    @staticmethod
    def _get_current_ip() -> IPv4Address:
        ip_checkers = IP_CHECKERS

        actual_ip = None
        while actual_ip is None:
            try:
                actual_ip = (
                    APIChecker(
                        url=f"https://{ip_checkers[0]}",
                        headers={"User-Agent": "curl/7.75"},
                        validation_func=lambda x: x,
                        ip_func=lambda x: x.strip(),
                    )
                    .run()
                    .actual_ip
                )
                return actual_ip

            except RequestException:
                ip_checkers = ip_checkers[1:]

    def run(self) -> EnsureVPNResult:
        actual_ip = IPChecker._get_current_ip()
        return EnsureVPNResult(
            is_connected=self.validation_func(actual_ip), actual_ip=actual_ip
        )



from datetime import datetime
from typing import Generator


def get_dict_values(key, d) -> Generator:
    """
    Returns all values for "key" in nested dict
    """
    if not d:
        return None
    if hasattr(d, "items"):
        for k, v in d.items():
            if k == key:
                yield v
            if isinstance(v, dict):
                for result in get_dict_values(key, v):
                    yield result
            elif isinstance(v, list):
                for d in v:
                    for result in get_dict_values(key, d):
                        yield result


def is_today(date: str, format: str = "%a %b %d %H:%M:%S %Y"):
    parsed_date = datetime.strptime(date, format)
    return parsed_date.date() == datetime.today().date()



from ipaddress import AddressValueError

from .exceptions import EnsureVPNException, VPNNotConnectedException
from .providers import CustomVPN, HideMyAssVPN, MullvadVPN, NordVPN, ProtonVPN

providers = [CustomVPN, HideMyAssVPN, MullvadVPN, NordVPN, ProtonVPN]


def ensure_vpn(ip_or_provider: str) -> None:
    try:
        selected_provider = CustomVPN(ip_or_provider)
    except AddressValueError:
        selected_providers = [
            p for p in providers if p.name.lower() == ip_or_provider.strip().lower()  # type: ignore
        ]
        if len(selected_providers) != 1:
            raise EnsureVPNException(
                f"No or too many VPN providers found (results: {selected_providers})"
            )

        selected_provider = selected_providers[0]  # type: ignore

    result = selected_provider.validate()
    if result.is_connected is False:
        raise VPNNotConnectedException(
            f"You are not connected to {selected_provider.name}. Found IP: {str(result.actual_ip)}",
            actual_ip=result.actual_ip,
        )


def ensure_vpn_decorator(provider: str):
    def _ensure_vpn_decorator(f):
        def wrapper(*args, **kwargs):
            ensure_vpn(provider)
            return f(*args, **kwargs)

        return wrapper

    return _ensure_vpn_decorator



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ensure_vpn']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'ensure-vpn',
    'version': '0.4.0',
    'description': 'A function to ensure you are connected to your favorite VPN before running your script.',
    'long_description': 'A Python function to ensure you are connected to your favorite VPN before running your script or function. It just raises an exception if you\'re not connected.\n\n## Supported VPN providers\n- Custom IP\n- HideMyAss(`"hidemyass"`)\n- Mullvad (`"mullvad"`)\n- NordVPN (`"nordvpn"`)\n- ProtonVPN (`"protonvpn"`)\n\nAdd your own!\n\n## Installation\n```\npip install ensure-vpn\n```\n\n## Usage\n\nImport the function and run it as the first thing in your script:\n\n```python\nfrom ensure_vpn import ensure_vpn\n\nensure_vpn("mullvad") # raises VPNNotConnectedException if you\'re not connected.\n\n# rest of your script goes here\n```\n\nYou can also use a custom IP or subnet:\n```python\nensure_vpn("2.235.200.110") # or e.g. "2.235.200.0/24"\n```\n\nYou can also use the decorator to run the check every time before running a specific function. This is to make sure you don\'t run untrusted code if you lose your VPN connection after starting your program.\n\nNote that this can be resource intensive depending on how often you call your function so it may slow down your program considerably or get you rate-limited by the services used by this script.\n\n```python\nfrom ensure_vpn import ensure_vpn_decorator\n\n@ensure_vpn_decorator("nordvpn")\ndef do_stuff():\n    # ...\n\ndo_stuff() # VPN is checked every time you call do_stuff\n```\n',
    'author': 'Francesco Truzzi',
    'author_email': 'francesco@truzzi.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ftruzzi/ensure_vpn',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
